/*
 *  Copyright (C) 2002 Peter Amstutz
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <stdarg.h>
#include <sys/types.h>
#include <stdlib.h>

#include "packet.h"

/* pack and unpack functions are from Kernighan & Pike's
   _The Practice of Programming_ chapter 9 pg 218,

   modified to pack length-specified strings as well (up to 64k
   characters long); also adds checks to against possible buffer
   overflows.
*/
int pack(pack_ubyte *buf, int maxsize, char *fmt, ...)
{
    va_list args;
    char* p;
    pack_ubyte* bp;
    pack_ubyte* st;
    pack_uint16 s, len;
    pack_uint32 l, i;

    bp=buf;
    va_start(args, fmt);
    for(p=fmt; *p; p++)
    {
        switch(*p)
        {
        case 'c':
            if((bp - buf + 1) > maxsize) return -1;
            ///*bp++ = (ub*bp++ = (pack_ubyte)va_arg(args, pack_ubyte); XXX gcc 2.96 complained about using pack_ubyte in va_arg
            *bp++ = (pack_ubyte)va_arg(args, int);
            break;
        case 's':
            //s = (pack_uint16)va_arg(args, pack_uint16); XXX gcc 2.86 complained about using pack_ubyte in va_arg
            if((bp - buf + 4) > maxsize) return -1;
            s = (pack_uint16)va_arg(args, int);
            *bp++ = s >> 8;
            *bp++ = s;
            break;
        case 'l':
            if((bp - buf + 4) > maxsize) return -1;
            l = (pack_uint32)va_arg(args, pack_uint32);
            *bp++ = l >> 24;
            *bp++ = l >> 16;
            *bp++ = l >> 8;
            *bp++ = l;
            break;
        case 'S':
            st = va_arg(args, pack_ubyte*);
            len = va_arg(args, int);
            if((bp - buf + 2 + len) > maxsize) return -1;
            *bp++ = len >> 8;
            *bp++ = len;
            for(i = 0; i < len; i++)
                *bp++ = *st++;
            break;
        default:
            va_end(args);
            return -1;
        }
    }
    va_end(args);
    return (bp - buf);
}

int unpack(pack_ubyte *buf, int maxsize, char *fmt, ...)
{
    va_list args;
    char *p;
    const pack_ubyte *bp;
    pack_ubyte *pc;
    pack_uint16 *ps;
    pack_uint32 *pl;
    pack_ubyte **st;
    int len, i;


    bp=buf;
    va_start(args, fmt);
    for(p=fmt; *p; p++)
    {
        switch(*p)
        {
        case 'c':
            if(bp - buf + 1 > maxsize) return -1;
            pc = va_arg(args, pack_ubyte*);
            *pc = *bp++;
            break;
        case 's':
            if(bp - buf + 2 > maxsize) return -1;
            ps = va_arg(args, pack_uint16*);
            *ps = *bp++ << 8;
            *ps |= *bp++;
            break;
        case 'l':
            if(bp - buf + 4 > maxsize) return -1;
            pl = va_arg(args, pack_uint32*);
            *pl = *bp++ << 24;
            *pl |= *bp++ << 16;
            *pl |= *bp++ << 8;
            *pl |= *bp++;
            break;
        case 'S':
            st = va_arg(args, pack_ubyte**);
            pl = va_arg(args, pack_uint32*);
            if(bp - buf + 2 > maxsize) return -1;
            len = *bp++ << 8;
            len |= *bp++;
            if(bp - buf + len > maxsize) return -1;
            *pl = len;
            *st = (pack_ubyte*)malloc(len+1);
            for(i = 0; i < len; i++)
                (*st)[i] = *bp++;
            (*st)[i] = 0;
            break;
        default:
            va_end(args);
            return -1;
        }
    }
    va_end(args);
    return bp - buf;
}
