/* $Id: service_responder.cc,v 1.9 2003/08/06 01:37:19 reed Exp $

    Copyright (C) 2002 Reed Hedges <reed@interreality.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <stdexcept>
#include <deque>
#include <set>
#include <iostream>
using namespace std;

#include "service_responder.hh"

extern void yyrestart(FILE*);

ServiceDirectory* serviceDir;
unsigned int launchDelay = 0;

#ifndef DEFAULT_CONFIG_FILE
#define DEFAULT_CONFIG_FILE "/etc/service_responder.conf"
#endif

void usage() {
    cerr << "usage: service_responder [-h] [-f <ConfigFile>] [-d <Delay>]" << endl;
    cerr << "    -h    Display this help message" << endl;
    cerr << "    -f    Load config file <ConfigFile> (default is " << DEFAULT_CONFIG_FILE << endl;
    cerr << "    -d    Sleep for <Delay> seconds after running a program" << endl;
}

int main(int argc, char** argv) {

    char *configfile = DEFAULT_CONFIG_FILE;
    char c;
    while( (c = getopt(argc, argv, "-hf:d:")) != -1) {
        switch(c) {
            case 'h':
                usage();
                exit(-1);
            case 'f':
                if(optarg)
                    configfile = optarg;
                else {
                    usage();
                    exit(-1);
                }
                break;
            case 'd':
                launchDelay = atoi(optarg);
                break;
            case 1:         // if random text is in argv, but not following -f
            case '?':
            default:
                cerr << "service_responder: Unrecognized option. Use -h for help.\n";
                exit(-1);
        }
    }

    // Create service directory object
    serviceDir = new ServiceDirectory();

    // Parse the config file. When it finds complete query/response rules 
    // in the cconfig file, then a quey will be made on serviceDir.
    assert(serviceDir);
    FILE* fp = fopen(configfile, "r");
    if(!fp) {
        cerr << "Error: could not open " << configfile << ":" << strerror(errno) << endl;;
        exit(-2);
    }
    yyrestart(fp);
    yyparse();
    fclose(fp);
    while(true)
        serviceDir->handleIncoming();
}


void Responder::notifyNewService(int id, const Service& s)
{
    int pid = fork();
    if(pid == 0) {
        char env_title[1024];
        char env_descrip[1024];
        char env_url[1024];
        char env_types[1024];

        sprintf(env_title, "SERVICE_TITLE=%s", s.title.c_str());
        putenv(env_title);
        sprintf(env_descrip, "SERVICE_DESCRIPTION=%s", s.description.c_str());
        putenv(env_descrip);
        sprintf(env_url, "SERVICE_URL=%s", s.url.c_str());
        putenv(env_url);

        string types("");
        for(set<string>::const_iterator i = s.types.begin(); i != s.types.end(); i++) {
            string t = *i;
            for(int c = 0; c < t.size(); c++) {
                if(t[c] == ',') {
                    t.replace(c, 1, "\\,"); 
                    c++;
                }
            }
            if(types == "")
                types = t;
            else
                types += "," + t;
        }
        sprintf(env_types, "SERVICE_TYPES=%s", types.c_str());
        putenv(env_types);

        char* cmd[args.size()+1];
        int ci = 0;
        for(deque<string>::const_iterator ai = args.begin(); ai != args.end();
    ai++)
            cmd[ci++] = strdup((*ai).c_str());
        cmd[ci] = NULL;
        execvp(cmd[0], cmd);
        cerr << "service_responder: error executing command \"" << cmd[0] << "\": " << strerror(errno) << endl;
        exit(-1);
    } else if (pid < 0) {
        cerr << "service_responder: error forking: " << strerror(errno) << endl;
    } else {
        cout << "service_responder: ran command with PID " << pid << " responding to service at " << s.url << endl;
    }

    sleep(launchDelay);
}

